
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#include <iostream>

#include "kernel/lib/kernel-learning-linearl1.h"
#include "kernel/lib/io-utilities.h"

using namespace kernel;

DEFINE_bool(sparse, false, "Use sparse matrix storage and operations.");
DEFINE_double(alg_reg, 1.0,
              "Regularization paremeter used for learning algorithm\n"
              "Used only with the lin1 method.");
DEFINE_double(ker_reg, 1.0,
              "Regularization parameter used for the kernel selection.");
DEFINE_double(tol, 1e-5, "Stopping criteria for lin1 method.");
DEFINE_int32(max_iter, 50, "Maximum number of itereation used for lin1.");
DEFINE_string(mu_file, "", "Save weights to output file.");
DEFINE_string(lk_alg, "unif",
             "\tunif -- uniform weights (i.e. standard linear kernel)\n"
             "\tlin1 -- l1-regularized linear combination");
DEFINE_string(num_train, "0:0",
        "Use the points in the range N:M (i.e. N, N+1, ..., M for training.\n"
        "0:0 indicated using all the points");


template <class M>
int klcombinefeatures(const vector<M> &data_vec,
                      const DenseVector<typename M::ElementType> &labels,
                      const char* outfile) {

  typedef typename M::ElementType T;

  // Select training points
  vector<M> data_train_vec;
  DenseVector<T> labels_train;

  // parse num_train flag
  vector<char*> tokens;
  size_t train_start, train_end;
  char *cstr = new char[FLAGS_num_train.size()+1];
  strcpy(cstr, FLAGS_num_train.c_str());
  SplitToVector(cstr, ":", &tokens, true);
  if (tokens.size() != 2)
    LOG(FATAL) << "Format error in num_train flag";
  train_start = atoi(tokens[0]);
  train_end = atoi(tokens[1]);
  delete[] cstr;

  if (train_start < train_end) {
    DenseVector<size_t> rows;
    //MakeRange(train_start, train_end, &rows);
    MakeRangeWrap(train_start, train_end, data_vec[0].NumRow(), &rows);
    LOG(INFO) << "Selecting " << rows.Size()
              << " training examples...";
    for (size_t i = 0; i < data_vec.size(); ++i) {
      M temp;
      SubMatrixCopy(data_vec[i], rows, &temp);
      data_train_vec.push_back(temp);
    }
    //SubsetCopy(train_start, train_end, labels, &labels_train);
    SubsetCopy(labels, rows, &labels_train);
  } else {
    data_train_vec = data_vec; //TODO(rostami): should not have to copy data
    labels_train = labels;
  }


  // Calculate weights
  // NOTE: all methods right now are regularized so that Sum(weights) = 1.0
  DenseVector<T> weights;

  if (FLAGS_lk_alg.compare("unif") == 0) {
    DenseVector<T> uniform(data_train_vec.size(), 1.0 / data_train_vec.size());
    weights = uniform;

  } else if (FLAGS_lk_alg.compare("lin1") == 0) {
    DenseVector<double> svm_vars;
    double offset;
    MakeLinearL1Weights(data_train_vec, true, labels_train, FLAGS_alg_reg,
                        FLAGS_tol, FLAGS_max_iter, &weights, &svm_vars,
                        &offset);

  } else {
    LOG(ERROR) << "Unknown weight type" << endl;
    return 1;
  }

  // Save weights if requested
  if (FLAGS_mu_file != "") {
    ofstream out(FLAGS_mu_file.c_str());
    if(!out.is_open()) {
      LOG(ERROR) << "klcombinefeatures: Unable to save weights" << endl;
    } else {
      PrintVector(weights, out);
      out.close();
    }
  }

  // Concatenate and print final features
  M features;
  DenseVector<T> weights_sqrt(weights.Size(), 0);
  for (size_t i = 0; i < weights.Size(); ++i)
    weights_sqrt.SetElement(i, sqrt(weights.Element(i)));
  ConcatenateWeightedFeatures(data_vec, weights_sqrt, &features);

  ofstream out(outfile);
  if(!out.is_open()) {
    LOG(ERROR) << "Error: Unable to output features" << endl;
  } else {
    FeaturePrintLibSVM(features, labels, out);
    out.close();
  }

  return 0;
}

int main(int argc, char** argv) {
  string usage =
      "Print concatenated features after weighting each set"
      "according to the selected automatic kernel selection method.\n\n";
  usage += "Usage: " + string(argv[0]);
  usage += " feature.list outfile\n";

  std::set_new_handler(FailedNewHandler);
  SetFlags(usage.c_str(), &argc, &argv, true);

  if (argc < 3) {
    ShowUsage();
    return 1;
  }

  vector<string> filenames;

  string line;
  ifstream file(argv[1]);
  if(!file.is_open()) {
    LOG(ERROR) << "Error: Could not read feature filename list: "
         << argv[1] << "." << endl;
    return 1;
  }
  while(1) {
    getline(file, line);
    if(file.eof()) break;
    filenames.push_back(line);
  }
  file.close();

  DenseVector<double> labels;
  int flag;
  if (FLAGS_sparse) {
    vector< SemiDenseMatrix<SparseVector<double>, DENSE_BY_ROW> > data_vec;

    for (size_t i = 0; i < filenames.size(); ++i) {
      SemiDenseMatrix<SparseVector<double>, DENSE_BY_ROW> data;
      flag = LibSVMToSparseMatrix(filenames[i], 0.0, &data, &labels);
      if (flag != 0) {
        LOG(ERROR) << "Error reading input file: " << filenames[i]
                   << "." << endl;
        LOG(ERROR) << "Code: " << flag << endl;
        return -1;
      }
      data_vec.push_back(data);
    }
    return klcombinefeatures(data_vec, labels, argv[2]);

  } else {
    vector< DenseMatrix<double> > data_vec;

    for (size_t i = 0; i < filenames.size(); ++i) {
      DenseMatrix<double> data;
      flag = LibSVMToDenseMatrix(filenames[i], 0.0, &data, &labels);
      if(flag != 0) {
        LOG(ERROR) << "Error reading input file: " << filenames[i]
                   << "." << endl;
        LOG(ERROR) << "Code: " << flag << endl;
        return -1;
      }
      data_vec.push_back(data);
    }
    return klcombinefeatures(data_vec, labels, argv[2]);
  }
}
