
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#include "kernel/lib/io-utilities.h"
#include "kernel/lib/krr.h"

DEFINE_bool(kernel, false, "Data argument is a precomputed kernel in\n"
            "libSVM format.");
DEFINE_bool(kar, false, "Data argument is an OpenKernel kar object.");
DEFINE_bool(sparse, false, "Store data in sparse format.");

using namespace kernel;

double RMSE(const DenseVector<double> &v1,
            const DenseVector<double> &v2) {
  CHECK(v1.Size() == v2.Size());
  double rmse = 0;
  for (size_t i = 0; i < v1.Size(); ++i)
    rmse += pow((v1.Element(i) - v2.Element(i)), 2);

  rmse = sqrt(rmse / v1.Size());
  return rmse;
}

int ReadModel(const string &filename,
              DenseVector<size_t> *ids,
              DenseVector<double> *params,
              bool *is_dual,
              bool *is_centered) {
  ifstream file(filename.c_str());
  if (!file.is_open())
    return 1;

  string line;
  getline(file, line);
  if (line == "dual_vars")
    *is_dual = true;
  else if (line == "primal_vars")
    *is_dual = false;
  else
    return 2;

  getline(file, line);
  if (line == "centered")
    *is_centered = true;
  else if (line == "not_centered")
    *is_centered = false;
  else
    return 3;

  vector<char*> tokens;
  char *cstr;
  size_t i = 0;
  ids->Resize(1,0);
  params->Resize(1,0);
  // TODO(Afshin): line != NULL, what was that?
  // while ((getline(file, line)) && line != NULL) {
  while ((getline(file, line)) && !line.empty()) {
    cstr = new char[line.size() + 1];
    strcpy(cstr, line.c_str());
    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);
    if (tokens.size() < 2) {
      delete[] cstr;
      return 4;
    }

    ids->AddElement(i, atoi(tokens[0]));
    params->AddElement(i, strtod(tokens[1], NULL));
    delete[] cstr;
    ++i;
  }

  file.close();
  return 0;
}

int ReadTestKernel(const string &filename,
                   const DenseVector<size_t> &ids,
                   DenseMatrix<double> *kernel,
                   DenseVector<double> *labels) {
  string line;
  vector<char*> tokens;
  char *cstr;
  size_t i;

  //FIRST PASS: Store labels and count number of test points
  ifstream file(filename.c_str());
  if(!file.is_open())
    return 1;
  labels->Resize(1, 0);
  i = 0;
  // TODO(Afshin): line != NULL, what was that?
  // while ((getline(file, line)) && line != NULL) {
  while ((getline(file, line)) && !line.empty()) {
    cstr = new char[line.size() + 1];
    strcpy(cstr, line.c_str());
    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);
    labels->AddElement(i, strtod(tokens[0], NULL));
    delete[] cstr;
    ++i;
  }
  file.close();

  kernel->Reset(labels->Size(), ids.Size(), 0);
  //SECOND PASS: fill in kernel matrix
  file.open(filename.c_str());
  if(!file.is_open())
    return 1;
  i = 0;
  size_t id;
  vector<char*> subtokens;

  // TODO(Afshin): line != NULL, what was that?
  // while ((getline(file, line)) && line != NULL) {
  while ((getline(file, line)) && !line.empty()) {
    cstr = new char[line.size() + 1];
    strcpy(cstr, line.c_str());
    tokens.clear();
    SplitToVector(cstr, " \t", &tokens, true);
    if(tokens.size() + 2 < ids.Element(ids.Size() - 1)) {
      delete[] cstr;
      return 1;
    }

    for (size_t j = 0; j < ids.Size(); ++j) {
      subtokens.clear();
      SplitToVector(tokens[ids.Element(j) + 1], ":", &subtokens, true);
      if((subtokens.size() < 2)
         || ((id = atoi(subtokens[0])) != ids.Element(j))) {
        delete[] cstr;
        return 1;
      }
      kernel->SetElement(i, j, strtod(subtokens[1], NULL));
    }

    delete[] cstr;
    ++i;
  }

  file.close();
  return 0;
}

int main(int argc, char** argv) {
  string usage =
      "Predict using KRR model generated by krr-train.\n\n";
  usage += "Usage: " + string(argv[0]);
  usage += " test_data krr_model [predictions]\n";

  std::set_new_handler(FailedNewHandler);
  SetFlags(usage.c_str(), &argc, &argv, true);

  if (argc < 3) {
    ShowUsage();
    return 1;
  }

  string outfile;
  if (argc >= 4) {
    outfile = argv[3];
  } else {
    outfile = argv[1];
    outfile += ".pred";
  }

  DenseVector<size_t> ids;
  DenseVector<double> params, labels, preds;
  bool is_dual, is_centered;
  int read_flag;
  if ((read_flag = ReadModel(argv[2], &ids, &params,
                            &is_dual, &is_centered))) {
    LOG(ERROR) << "ERROR: Cannot read model file." << endl;
    LOG(ERROR) << "error code: " << read_flag << endl;
    return 1;
  }

  //kernel case
  if (is_dual) {
    LOG(INFO) << "Using dual solution to make predictions...";
    DenseMatrix<double> test_kernel(1, 1, 0);
    if (FLAGS_kar) {
      LOG(ERROR) << "--kar feature not yet implemented." << endl;
    } else if (FLAGS_kernel) {
      if (ReadTestKernel(argv[1], ids, &test_kernel, &labels)) {
        LOG(ERROR) << "ERROR: Cannot read kernel file." << endl;
        return 1;
      }
    } else {
      LOG(ERROR) << "ERROR: Must supply a kernel to make predictions";
      return 1;
    }
    LOG(INFO) << "Making predicitons...";
    KRRDualPredict(test_kernel, params, &preds);
  //feature map case
  } else { //sparse subcase
    LOG(INFO) << "Using primal solution to make predictions...";
    if(FLAGS_sparse) {
      SemiDenseMatrix<SparseVector<double>, DENSE_BY_ROW> data_sparse;
      if(LibSVMToSparseMatrix(argv[1], &data_sparse, &labels) != 0) {
        LOG(ERROR) << "Error: Problem reading sparse data file.";
        return 1;
      }
      KRRPrimalPredict(data_sparse, params, &preds);
    } else { //dense subcase
      DenseMatrix<double> data_dense;
      if(LibSVMToDenseMatrix(argv[1], &data_dense, &labels) != 0) {
        LOG(ERROR) << "Error: Problem reading sparse data file.";
        return 1;
      }
      KRRPrimalPredict(data_dense, params, &preds);
    }
  }

  double rmse = RMSE(preds, labels);
  LOG(INFO) << "RMSE: " << rmse;

  ofstream out(outfile.c_str());
  if (!out.is_open()) {
    LOG(INFO) << "ERROR: could not open output file";
    return 1;
  }
  for (size_t i = 0; i < preds.Size(); ++i)
    out << preds.Element(i) << endl;
  out.close();

  return 0;
}
