// dataset.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Dataset representation.

#ifndef KERNEL_LIB_DATASET_H__
#define KERNEL_LIB_DATASET_H__

#include "kernel/lib/object.h"
#include "kernel/lib/register.h"
#include <fst/extensions/far/sttable.h>
#include <fst/util.h>

#include <unistd.h>
#include <sys/param.h>

namespace kernel {

struct KernelReadOptions;
struct KernelWriteOptions;

using namespace fst;

class DataSetBase {
 public:
  virtual ~DataSetBase() {}
  virtual size_t Size() const = 0;
  virtual const string &File() const = 0;
};

template <class O> class DataSet : public DataSetBase {
 public:

  DataSet(const vector<O*> &vec) : objects_(vec), ref_count_(1) {}

  DataSet(const string &filename) : ref_count_(1) {
    if (filename[0] != '/') {
      char path[PATH_MAX];
      file_ = string(getcwd(path, PATH_MAX)) + '/' + filename;
    } else {
      file_ = filename;
    }

    ifstream is(file_.c_str());
    if (!is)
      LOG(FATAL) << "DataSet: cannot open dataset file " << file_ << endl;

    int32 magic_number = 0;
    ReadType(is, &magic_number);
    is.seekg(0);

    if (magic_number == kSTTableMagicNumber) {
      vector<string> filenames;
      filenames.push_back(file_);
      STTableReader<O, ObjectRead<O> > reader(filenames);
      while (!reader.Done()) {
        O *obj = ObjectCopy<O>()(reader.GetEntry());
        objects_.push_back(obj);
        reader.Next();
      }
    } else {
      string s;
      size_t nline = 0;
      while (getline(is, s)) {
        if (s.empty()) break;
        O *obj = ObjectReadText<O>()(s, filename, ++nline);
        if (!obj)
          LOG(FATAL) << "DataSet: cannot read object '" << s << "' from "
                     << "dataset file: " << file_ << ", line: " << nline
                     << endl;
        objects_.push_back(obj);
      }
    }
  }

  ~DataSet() {
    VLOG(1) << "~DataSet[" << RefCount() << "]";
    if (!DecrRefCount()) {
      for (size_t i = 0; i < objects_.size(); ++i)
        delete objects_[i];
      DataRegister *registr = DataRegister::GetRegister();
      if (registr->GetData(file_) == this)
        registr->SetData(file_, 0);
    }
  }

  size_t Size() const { return objects_.size();}

  const O &operator[](size_t i) const { return *(objects_[i]);}

  static DataSet<O> *Read(istream &strm, const KernelReadOptions &opts);

  static DataSet<O> *Read(const string &filename) {
    ifstream strm(filename.c_str());
    if (!strm) {
      LOG(ERROR) << "DataSet::Read: Can't open file: " << filename;
      return 0;
    }
    return Read(strm, KernelReadOptions(filename));
  }

  bool Write(ostream &strm, const KernelWriteOptions &opts) const;

  bool Write(const string &filename) const {
    if (!filename.empty()) {
      ofstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "DataSet::Write: Can't open file: " << filename;
        return false;
      }
      return Write(strm, KernelWriteOptions(filename));
    } else {
      return Write(std::cout, KernelWriteOptions("standard output"));
    }
  }

  int RefCount() { return ref_count_;}

  int IncrRefCount() { return ++ref_count_;}

  int DecrRefCount() { return --ref_count_;}

  const string &File() const {return file_;}

 private:
  vector<O*> objects_;
  string file_;
  int ref_count_;
};


template <class O>
bool DataSet<O>::Write(ostream &strm, const KernelWriteOptions &opts) const {
  return WriteType(strm, file_);
}

template <class O>
DataSet<O>* DataSet<O>::Read(istream &strm, const KernelReadOptions &opts){
  string s;
  ReadType(strm, &s);
  DataRegister *registr = DataRegister::GetRegister();
  DataSetBase *data = registr->GetData(s);
  if (!data) {
    data = new DataSet<O>(s);
    registr->SetData(s, data);
  }
  return down_cast<DataSet<O>*>(data);
}


}  // namespace kernel

#endif  // KERNEL_LIB_DATASET_H__
