
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_KERNEL_LEARNING_LINEARL1_H_
#define KERNEL_LIB_KERNEL_LEARNING_LINEARL1_H_

#include <cmath>

#include <glpk.h>
#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/krr.h"
#include "kernel/lib/learn-correlation.h"
#include "kernel/lib/kernel-learning-linearl2.h"
#include "kernel/lib/liblinear.h"
#include "kernel/lib/libsvm.h"
#include "kernel/lib/quadprog.h"

// This file includes the implementation of an iterative algorithm to solve
// the semi-infinite linear program suggested by Sonnenburg et al. (JMLR 2006).
// We alternate between solving an LP (via glpk) and an SVM (via LIBLINEAR or
// LIBSVM).

namespace kernel {

//
// HELPER FUNCTIONS
//

// Given an array of explicit feature mappings for a single dataset,
// their labels and the dual variables for an SVM problem, return the
// array of SVM objective values with respect to each feature mapping.
template <typename T, class M> inline
void GetSVMObjectives(const vector<M> &data, bool explicit_features,
                      const DenseVector<T> &labels,
                      const DenseVector<T> &alpha, DenseVector<T> *S) {

  typename M::TransposeMatrix data_t;
  DenseVector<T> signed_alpha(alpha.Size(), 0);
  S->Resize(data.size(), 0);

  double alpha_sum = 0;
  for (size_t i = 0; i < alpha.Size(); ++i) {
    alpha_sum += alpha.Element(i);
    signed_alpha.SetElement(i, alpha.Element(i) * Sign(labels.Element(i)));
  }

  DenseVector<double> temp;

  if (explicit_features) {
    for (size_t i = 0; i < data.size(); ++i) {
      data[i].Transpose(&data_t);
      Product(data_t, signed_alpha, &temp);
      S->SetElement(i, DotProduct(temp, temp));
    }

  } else { // data contains kernel matrices
    for (size_t i = 0; i < data.size(); ++i) {
      Product(data[i], signed_alpha, &temp);
      S->SetElement(i, DotProduct(signed_alpha, temp));
    }
  }

  for (size_t i = 0; i < S->Size(); ++i)
    S->SetElement(i, 0.5 * S->Element(i) - alpha_sum);
}

// Same as above, but each "kernel" is based on a single feature.
template <typename T, class M> inline
void GetSVMObjectives(const M &data, const DenseVector<T> &labels,
               const DenseVector<T> &alpha,
               DenseVector<T> *S) {
  typename M::TransposeMatrix data_t;
  data.Transpose(&data_t);
  DenseVector<T> signed_alpha(data.NumRow(), 0);
  S->Resize(data.NumColumn(), 0);

  double alpha_sum = 0;
  for (size_t i = 0; i < alpha.Size(); ++i) {
    alpha_sum += alpha.Element(i);
    signed_alpha.SetElement(i, alpha.Element(i) * Sign(labels.Element(i)));
  }

  Product(data_t, signed_alpha, S);
  for (size_t i = 0; i < S->Size(); ++i)
    S->SetElement(i, 0.5 * pow(S->Element(i), 2) - alpha_sum);
}

template <typename T> inline
T Sign(const T &x) {
  if (x > 0)
    return 1.0;
  else
    return -1.0;
}

// Initialize the lp structure and constraint coefficients, returns
// the index of the end of the current constraint matrix.
inline int GlpkSetup(size_t num_kernel, size_t max_iter,
                      glp_prob **lp, int **ai, int **aj, double **ac) {
  //glpk setup, variables are in order [theta, beta_1, ..., beta_n]
  size_t ic = 1; //index_counter
  *ai = new int[(num_kernel + 1) * (max_iter + 1)];
  *aj = new int[(num_kernel + 1) * (max_iter + 1)];
  *ac = new double[(num_kernel + 1) * (max_iter + 1)];
  *lp = glp_create_prob();
  glp_set_obj_dir(*lp, GLP_MAX);
  glp_add_cols(*lp, num_kernel + 1);
  glp_set_obj_coef(*lp, 1, 1.0); //maximize theta
  glp_set_col_bnds(*lp, 1, GLP_FR, 0.0, 0.0);
  // beta >= 0
  for (size_t i = 2; i <= num_kernel + 1; ++i)
    glp_set_col_bnds(*lp, i, GLP_LO, 0.0, 0.0);
  // sum_i beta_i = kernel_reg
  glp_add_rows(*lp, 1);
  glp_set_row_bnds(*lp, 1, GLP_FX, 1, 1);
  for (size_t i = 2; i <= num_kernel + 1; ++i) {
    (*ai)[ic] = 1;
    (*aj)[ic] = i;
    (*ac)[ic] = 1.0;
    ++ic;
  }
  return ic;
}

// Add a new violated constraint, returns the index of the end of the
// current constraint matrix.
template <typename T> inline
int GlpkAddConstraint(const DenseVector<T> &beta_coeffs, int ic,
                       glp_prob **lp, int **ai, int **aj, double **ac) {
    int row_id = glp_add_rows(*lp, 1);
    glp_set_row_bnds(*lp, row_id, GLP_LO, 0.0, 0.0);
    (*ai)[ic] = row_id;
    (*aj)[ic] = 1;
    (*ac)[ic] = -1.0; // -theta
    ++ic;
    // sum_i beta_i S_i
    for (size_t i = 2; i <= beta_coeffs.Size() + 1; ++i) {
      (*ai)[ic] = row_id;
      (*aj)[ic] = i;
      (*ac)[ic] = beta_coeffs.Element(i - 2);
      ++ic;
    }

    return ic;
}

// Solve lp, populate beta with optimal kernel weights.  Returns
// objective value of lp.
template <typename T> inline
double GlpkSolve(int ic, const int *ai, const int *aj,
               const double *ac, int num_kernel, glp_prob **lp,
               DenseVector<T> *beta) {
    if(beta->Size() != num_kernel)
      beta->Resize(num_kernel, 0);

    glp_load_matrix(*lp, ic-1, ai, aj, ac);
    if(glp_simplex(*lp, NULL) != 0)
      LOG(WARNING) << "Warning: LP did not terminate successfuly" << endl;
    double theta = glp_get_col_prim(*lp, 1);
    int nnz = 0;
    for (size_t i = 2; i <= num_kernel + 1; ++i) {
      beta->SetElement(i-2, glp_get_col_prim(*lp, i));
      if (beta->Element(i-2) > 0)
        ++nnz;
    }
    VLOG(1) << "GlpkSolve: number of non-zeros: " << nnz << endl;
    return theta;
}

// Given several explicit feature representations of the same data,
// concatenate the features after weighting each set.
template <class T> inline
void ConcatenateWeightedFeatures(
    const vector< SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> > &data,
    const DenseVector<T> &weights,
    SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> *weighted_data) {
  CHECK(data.size() == weights.Size());
  weighted_data->Clear();

  int *num_features = new int[data.size()]; //cumulative number of features
  for (size_t i = 0; i < data.size(); ++i) {
    CHECK(data[0].NumRow() == data[i].NumRow());
    num_features[i] = 0;
    for (size_t j = 0; j <= i; ++j)
      num_features[i] += data[j].NumColumn();
  }

  SparseVector<T> row;
  int offset;
  for (size_t i = 0; i < data[0].NumRow(); ++i) {
    offset = 0;
    SparseVector<T> new_row;
    for (size_t j = 0; j < data.size(); ++j) {
      row = data[j].Row(i);
      if(j > 0) offset = num_features[j-1];
      for (size_t k = 0; k < row.Size(); ++k)
        new_row.AddElement(row.Pair(k).first + offset,
                           row.Pair(k).second * weights.Element(j));
    }
    weighted_data->AddRow(new_row);
  }

  CHECK(weighted_data->NumRow() == data[0].NumRow());
  delete[] num_features;
}

// Given several explicit feature representations of the same data,
// concatenate the features after weighting each set.
template <class T> inline
void ConcatenateWeightedFeatures(const vector< DenseMatrix<T> > &data,
                                 const DenseVector<T> &weights,
                                 DenseMatrix<T> *weighted_data) {
  CHECK(data.size() == weights.Size());

  int *num_features = new int[data.size()]; //cumulative number of features
  for (size_t i = 0; i < data.size(); ++i) {
    CHECK(data[0].NumRow() == data[i].NumRow());
    num_features[i] = 0;
    for (size_t j = 0; j <= i; ++j)
      num_features[i] += data[j].NumColumn();
  }

  weighted_data->Reset(data[0].NumRow(), num_features[data.size()-1], 0);
  int offset;
  for (size_t i = 0; i < data[0].NumRow(); ++i) {
    offset = 0;
    for (size_t j = 0; j < data.size(); ++j) {
      if(j > 0) offset = num_features[j-1];
      for (size_t k = 0; k < data[j].NumColumn(); ++k)
        weighted_data->SetElement(i, k + offset,
                           weighted_data->Element(i, k) * weights.Element(j));
    }
  }

  CHECK(weighted_data->NumRow() == data[0].NumRow());
  delete[] num_features;
}

// MAIN FUNCTIONS

// Compute L1 regularized weights via the explicit feature representation,
// for each kernel.
// INPUT:
// data - a vector of Dense or SemiDense matrices, each being a feature
//   mapping of the same dataset. Each row represents an instance, thus it
//   should be the case that data[i].NumRow() == data[j].NumRow() for all i,j.
// explicit_features - if true, data contains feature mappings, if false
//   data constain kernel matrices.
// labels - a DenseVector of the labels of each instance.
// alg_reg - the regularization parameter that will be used with the SVM.
//   This shoudl match the regularization used when finally conducting
//   classification.
// stop_tol - the condition used to stop iterating, smaller values lead to
//   more precise solutions but convergence is more difficult.
//   (suggested value 1e-5).
// max_iter - the maximum number of iterations, will stop even if stop_tol
//   is not met.
// OUTPUT:
// weights - the optimal kernel combination weights.
// svm_vars - the optimal dual variables for the svm problem trained using
//   the weighted kernel combination.
//
//   For algorithm reference see: Sonnenburg et al. JMLR 2006
template <class M>
double MakeLinearL1Weights(const vector<M> &data,
                         bool explicit_features,
                         const DenseVector<typename M::ElementType> &labels,
                         double alg_reg,
                         double stop_tol,
                         size_t max_iter,
                         DenseVector<typename M::ElementType> *weights,
                         DenseVector<typename M::ElementType> *svm_vars,
                         double *offset) {
  typedef typename M::ElementType T;

  *offset = 0; //liblinear does not touch this

  //Variable names chosen to match Sonnenburg et al. paper.
  DenseVector<T> beta(data.size(), 1.0 / data.size());
  DenseVector<T> beta_sqrt(beta.Size(), 0);
  DenseVector<T> alpha(data[0].NumRow(), 0);
  DenseVector<T> diff;
  DenseVector<T> S;
  M data_weighted;

  size_t iter = 0;
  double constraint = 100.0, theta = 1.0, gap = 1e100;

  //glpk setup, variables are in order [theta, beta_1, ..., beta_n]
  glp_prob *lp;
  int *ai, *aj;
  double *ac;
  size_t ic = GlpkSetup(data.size(), max_iter, &lp, &ai, &aj, &ac);

  while (iter < max_iter) {

    if (explicit_features) {
      //weight and concatenate data, for single kernel problem
      for (size_t i = 0; i < beta.Size(); ++i)
        beta_sqrt.SetElement(i, sqrt(beta.Element(i)));
      ConcatenateWeightedFeatures(data, beta_sqrt, &data_weighted);
  
      //Find minimizing alpha
      LibLinearTrain(data_weighted, labels, alg_reg, &alpha);

    } else { //data contains kernel matrices
      //weight and concatenate data, for single kernel problem
      LinearCombination(data, beta, &data_weighted);

      //Find minimizing alpha
      LibSVMTrain(data_weighted, labels, alg_reg, &alpha, offset);
    }

    //add new constraint sum_i beta_i S_i - theta >= 0
    GetSVMObjectives(data, explicit_features, labels, alpha, &S);
    ic = GlpkAddConstraint(S, ic, &lp, &ai, &aj, &ac);

    constraint = DotProduct(S, beta);
    gap = 1.0 - constraint / theta;
    gap *= Sign(gap);
    LOG(INFO) << "iter: " << iter
              << " constraint: " << constraint
              << " theta: " << theta
              << " gap: " << gap;
    if (gap < stop_tol)
      break;

    //Find maximizing theta & beta
    theta = GlpkSolve(ic, ai, aj, ac, data.size(), &lp, &beta);
    ++iter;
  }

  delete[] ai;
  delete[] aj;
  delete[] ac;
  glp_delete_prob(lp);

  *weights = beta;
  *svm_vars = alpha;
  return gap;
}


// Compute L1 regularized weights via the explicit feature representation,
// using each feature as a kernel.
//
// The parameters here are the same as the above function with the
// exception of data, which is a single feature representation of the
// data.  The algorithm then produces a weight for each feature.
//
// This is still EXPERIMENTAL -- known not to converge when using many
// features.
template<class M>
double MakeLinearL1Weights(const M &data,
                         const DenseVector<typename M::ElementType> &labels,
                         double alg_reg,
                         double stop_tol,
                         size_t max_iter,
                         DenseVector<typename M::ElementType> *weights,
                         DenseVector<typename M::ElementType> *svm_vars) {
  typedef typename M::ElementType T;

  //Variable names chosen to match Sonnenburg et al. paper.
  DenseVector<T> beta(data.NumColumn(), 1.0/data.NumColumn());
  DenseVector<T> beta_sqrt(data.NumColumn(), 0);
  DenseVector<T> alpha(data.NumRow(), 0);
  DenseVector<T> diff;
  DenseVector<T> S;
  M data_weighted;

  size_t iter = 0;
  double constraint = 100.0, theta = 1.0, gap = 1e100;

  //glpk setup, variables are in order [theta, beta_1, ..., beta_n]
  glp_prob *lp;
  int *ai, *aj;
  double *ac;
  size_t ic = GlpkSetup(data.NumColumn(), max_iter,
                        &lp, &ai, &aj, &ac);

  while (iter < max_iter) {

    //weight data, for single kernel problem
    for (size_t i = 0; i < beta.Size(); ++i)
      beta_sqrt.SetElement(i, sqrt(beta.Element(i)));
    DiagonalProduct(data, beta_sqrt, &data_weighted);

    //Find minimizing alpha
    LibLinearTrain(data_weighted, labels, alg_reg, &alpha);

    //add new constraint sum_i beta_i S_i - theta >= 0
    GetSVMObjectives(data, labels, alpha, &S);
    ic = GlpkAddConstraint(S, ic, &lp, &ai, &aj, &ac);

    constraint = DotProduct(S, beta);
    gap = 1.0 - constraint / theta;
    gap *= Sign(gap);
    LOG(INFO) << "iter: " << iter
              << " constraint: " << constraint
              << " theta: " << theta
              << " gap: " << gap;
    if (gap < stop_tol)
      break;

    //Find maximizing theta & beta
    theta = GlpkSolve(ic, ai, aj, ac, data.NumColumn(), &lp, &beta);
    ++iter;
  }

  delete[] ai;
  delete[] aj;
  delete[] ac;
  glp_delete_prob(lp);

  *weights = beta;
  *svm_vars = alpha;
  return gap;
}

// Center the data K in the kernel space, according to the statistics
// (x_bar) in G.
//   K_centered(x_i,x_j)
//      = K(x_i,x_j) - K(x_i, x_bar) - K(x_j, x_bar) + K(x_bar, x_bar).
template <typename U, typename T>
void CenterKernel(const DenseVector<U> &train_points, DenseMatrix<T> *K) {
  DenseVector<U> all_points;
  MakeRange(0, K->NumColumn(), &all_points); 

  size_t m = train_points.Size();
  size_t n = all_points.Size();
  double Gbar = (1.0 / pow(m, 2.0)) * Sum(*K, train_points, train_points);
  double GKrow[n];
  DenseVector<U> row(1,0);
  for (size_t i = 0; i < n; ++i) {
    row.SetElement(0, i);
    GKrow[i] = (1.0 / m) * Sum(*K, row, train_points);
  }

  double temp;
  for (size_t i = 0; i < n; ++i)
    for (size_t j = 0; j < i; ++j) {
      temp = K->Element(i, j) - GKrow[i] - GKrow[j] + Gbar;
      K->SetElement(i, j, temp);
      K->SetElement(j, i, temp);
    }
  for (size_t i = 0; i < n; ++i)
    K->SetElement(i, i, K->Element(i,i) - 2 * GKrow[i] + Gbar);
}

// Center the data in the kernel space.
//   K_centered(x_i,x_j)
//      = K(x_i,x_j) - K(x_i, x_bar) - K(x_j, x_bar) + K(x_bar, x_bar).
template <typename T>
void CenterKernel(DenseMatrix<T> *K) {
  DenseVector<size_t> all_points;
  MakeRange(0, K->NumColumn(), &all_points); 
  
  CenterKernel(all_points, K);
}

// TODO: finish this, either in the dual form with beta's (as is started
//       now), or with a solver that can give dual variables.
template<class M>
double MakeLinearL1WeightsQP(const M &data,
                       const DenseVector<typename M::ElementType> &labels,
                       double alg_reg,
                       double ker_reg,
                       DenseVector<typename M::ElementType> *weights,
                       DenseVector<typename M::ElementType> *svm_vars) {

  size_t m = data.NumRow();
  size_t n = data.NumColumn();
  DenseMatrix<typename M::ElementType> H(n, n, 1 / (2 * ker_reg)), A, B;
  DenseVector<typename M::ElementType> f(n, alg_reg), b, c, u, l;

  DenseMatrix<typename M::ElementType> U, Ut;
  
  Transpose(&U);
  DiagonalProduct(U, labels, &U);
  Transpose(&U);

}


} //namespace kernel

#endif  // KERNEL_LIB_KERNEL_LEARNING_LINEARL1_H_
