// kernel.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Abstract base class for kernel representation.

#ifndef KERNEL_LIB_KERNEL_H__
#define KERNEL_LIB_KERNEL_H__

#include <cmath>
#include <string>

#include "kernel/lib/compat.h"

#include "kernel/lib/dataset.h"
#include "kernel/lib/register.h"
#include "kernel/lib/properties.h"


namespace kernel {

class KernelHeader;

struct KernelReadOptions {
  string source;
  const KernelHeader *header;
  string type;

  explicit KernelReadOptions(const string& src = "<unspecified>",
                             const KernelHeader *hdr = 0) :
      source(src), header(hdr), type() {}
};

struct KernelWriteOptions {
  string source;
  bool write_header;
  bool write_cache;

  explicit KernelWriteOptions(const string& src = "<unspecified>",
                              bool cache = true,
                              bool hdr = true):
      source(src), write_header(hdr), write_cache(cache) {}
};


//
// Kernel HEADER CLASS
//
// This is the recommended Kernel file header representation.
//
class KernelHeader {
 public:
  KernelHeader() : version_(0), flags_(0), properties_(0) {}
  const string &KernelType() const { return kerneltype_; }
  const string &ObjectType() const { return objecttype_; }
  int32 Version() const { return version_; }
  int32 GetFlags() const { return flags_; }
  uint64 Properties() const { return properties_; }

  void SetKernelType(const string& type) { kerneltype_ = type; }
  void SetObjectType(const string& type) { objecttype_ = type; }
  void SetVersion(int32 version) { version_ = version; }
  void SetFlags(int32 flags) { flags_ = flags; }
  void SetProperties(uint64 properties) { properties_ = properties; }

  bool Read(istream &strm, const string &source);
  bool Write(ostream &strm, const string &source) const;

 private:
  string kerneltype_;  // E.g. "rational"
  string objecttype_;  // E.g. "log_fst"
  int32 version_;      // File format version
  int32 flags_;
  uint64 properties_;  // Kernel property bits
};

//
// Kernel INTERFACE CLASS DEFINITION
//

// A generic abstract Kernel with common-denominator methods.
class KernelBase {
 public:
  virtual ~KernelBase() {}

  virtual const string &Type() const = 0;        // Kernel type name

  virtual const string &ObjectType() const = 0;  // Object type name

  // Compute the Kernel between objects of index i and j in the
  // dataset the Kernel is defined on.
  virtual double Eval(size_t i, size_t j) = 0;

  // Return stored properties fr mask.
  virtual uint64 Properties(uint64 mask = kKernelProperties) const = 0;

  // Set properties with respect to mask.
  virtual void SetProperties(uint64 props,
                             uint64 mask = kKernelProperties) = 0;

  // Get a copy of this Kernel
  virtual KernelBase *Copy() const = 0;

  // Read a Kernel from an input stream; return NULL on error
  static KernelBase *Read(istream &strm, const KernelReadOptions &opts) {
    KernelReadOptions ropts(opts);
    KernelHeader hdr;
    if (ropts.header)
      hdr = *opts.header;
    else {
      if (!hdr.Read(strm, opts.source))
        return 0;
      ropts.header = &hdr;
    }
    KernelRegister *registr = KernelRegister::GetRegister();
    const KernelRegister::Reader reader =
      registr->GetReader(hdr.ObjectType(), hdr.KernelType());
    if (!reader) {
      LOG(ERROR) << "KernelBase::Read: Unknown kernel type \""
                 << hdr.KernelType()
                 << "\" (object type = \"" << hdr.ObjectType()
                 << "\"): " << ropts.source;
      return 0;
    }
    return reader(strm, ropts);
  };

  // Read a Kernel from file; return NULL on error
  // Empty filename reads from standard input
  static KernelBase *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "Kernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

  // Write an Kernel to an output stream; return false on error
  virtual bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    LOG(ERROR) << "Fst::Kernel: No write method for " << Type()
               << " kernel type";
    return false;
  }

  // Write an Kernel to a file; return false on error
  virtual bool Write(const string &filename) const {
    LOG(ERROR) << "Fst::Write: No write method for "
               << Type() << " Fst type: "
               << (filename.empty() ? "standard output" : filename);
    return false;
  }

  // Return the dataset over which the Kernel is defined
  virtual const DataSetBase *Data() const = 0;
};


// A generic Kernel, templated on the object definition, with
// common-denominator methods.
template <class O>
class Kernel : public KernelBase {
 public:
  typedef O Object;

  virtual ~Kernel() {}

  const string &ObjectType() const { return kernel::ObjectType<O>::type;}

  virtual Kernel<O> *Copy() const = 0;

  virtual const DataSet<Object> *Data() const = 0;

  virtual DataSet<Object> *Data() = 0;

  virtual void SetData(DataSet<Object> *) = 0;


  static Kernel<O> *Read(const string &filename) {
    return down_cast<Kernel<O>*>(KernelBase::Read(filename));
  }

  static Kernel<O> *Read(istream &strm, const KernelReadOptions &opts) {
    return down_cast<Kernel<O>*>(KernelBase::Read(strm, opts));
  }
};

//
// Kernel IMPLEMENTATION BASE CLASS
//
// This is the recommended Kernel implementation base class. It will
// handle reference counts, property bits, type information and
// dataset.
//
template <class O>
class KernelImpl {
 public:
  typedef O Object;

  KernelImpl()
      : properties_(0), type_("null"), dataset_(0), ref_count_(1) {}

  KernelImpl(const KernelImpl &impl)
      : properties_(impl.properties_), type_(impl.type_),
        dataset_(impl.dataset_), ref_count_(1) {}

  ~KernelImpl() {
    VLOG(1) << "~KernelImpl[" << (dataset_ ? dataset_->RefCount() : -1) << "]";
    if (dataset_ && (dataset_->DecrRefCount() == 0))
      delete dataset_;
  }

  const string& Type() const { return type_; }

  void SetType(const string &type) { type_ = type; }

  uint64 Properties() const { return properties_; }

  uint64 Properties(uint64 mask) const { return properties_ & mask; }

  void SetProperties(uint64 props) { properties_ = props; }

  void SetProperties(uint64 props, uint64 mask) {
    properties_ &= ~mask;
    properties_ |= props & mask;
  }

  const DataSet<Object>* Data() const { return dataset_; }

  DataSet<Object>* Data() { return dataset_; }

  void SetData(DataSet<Object>* ds) {
    if (ds == dataset_) return;
    if (dataset_ && (dataset_->DecrRefCount() == 0))
      delete dataset_;
    if(ds) ds->IncrRefCount();
    dataset_ = ds;
  }

  int RefCount() const { return ref_count_; }

  int IncrRefCount() { return ++ref_count_; }

  int DecrRefCount() { return --ref_count_; }

  // Read-in header and symbols, initialize kernel, and return the header.
  // If opts.header is non-null, skip read-in and use the option value.
  bool ReadHeader(istream &strm, const KernelReadOptions& opts,
                  int min_version, KernelHeader *hdr) {
    if (opts.header)
      *hdr = *opts.header;
    else if (!hdr->Read(strm, opts.source))
      return false;
    if (hdr->KernelType() != type_) {
      LOG(ERROR) << "KernelImpl::ReadHeader: Kernel not of type \""
                 << type_ << "\": " << opts.source;
      return false;
    }
    if (hdr->ObjectType() != ObjectType<O>::type) {
      LOG(ERROR) << "KernelImpl::ReadHeader: Object not of type \""
                 << ObjectType<O>::type
                 << "\": " << opts.source;
      return false;
    }
    if (hdr->Version() < min_version) {
      LOG(ERROR) << "KernelImpl::ReadHeader: Obsolete "
                 << type_ << " kernel version: " << opts.source;
      return false;
    }
    properties_ = hdr->Properties();
    return true;
  }

  // Write-out header and symbols.
  // If a opts.header is false, skip writing header.
  void WriteHeader(ostream &strm, const KernelWriteOptions& opts,
                   int version, KernelHeader *hdr) const {
    if (opts.write_header) {
      hdr->SetKernelType(type_);
      hdr->SetObjectType(ObjectType<O>::type);
      hdr->SetVersion(version);
      hdr->SetProperties(properties_);
      int32 file_flags = 0;
      hdr->SetFlags(file_flags);
      hdr->Write(strm, opts.source);
    }
  }


 protected:
  uint64 properties_;           // Property bits

 private:
  string type_;                 // Unique name of Kernel class
  DataSet<Object> *dataset_;    // Data set
  int ref_count_;               // Reference count

  void operator=(const KernelImpl<O> &impl);  // disallow
};


}  // namespace kernel

#endif  // KERNEL_LIB_KERNEL_H__
