// product-kernel.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Product kernel representation.

#ifndef KERNEL_LIB_PRODUCT_KERNEL_H__
#define KERNEL_LIB_PRODUCT_KERNEL_H__


#include "kernel/lib/kernel.h"

namespace kernel {


template <class O>
class ProductKernelImpl : public KernelImpl<O> {
 public:
  typedef O Object;

  ProductKernelImpl(const Kernel<O> &k1, const Kernel<O> &k2)
      : kernel1_(k1.Copy()), kernel2_(k2.Copy()) {
    KernelImpl<O>::SetType("product");
    Init();
    KernelImpl<O>::SetProperties(
        ProductProperties(kernel1_->Properties(), kernel2_->Properties()));
  }

  ~ProductKernelImpl() {
    delete kernel1_;
    delete kernel2_;
  }

  double Eval(size_t i, size_t j) {
    return kernel1_->Eval(i, j) * kernel2_->Eval(i,j);
  }

  void SetData(DataSet<O> *ds) {
    kernel1_->SetData(ds);
    kernel2_->SetData(ds);
    KernelImpl<O>::SetData(ds);
  }

  static ProductKernelImpl<O> *Read(istream &strm,
                                    const KernelReadOptions &opts) {
    ProductKernelImpl<O> *impl = new ProductKernelImpl<O>();
    KernelHeader hdr;
    if (!impl->ReadHeader(strm, opts, kMinFileVersion, &hdr))
        return 0;
    Kernel<O> *kernel1 = down_cast<Kernel<O>*>(
        Kernel<O>::Read(strm, KernelReadOptions(opts.source)));
    if (!kernel1) {
      LOG(ERROR) << "ProductKernel::Read: read failed: " << opts.source;
      return 0;
    }
    Kernel<O> *kernel2 = down_cast<Kernel<O>*>(
        Kernel<O>::Read(strm, KernelReadOptions(opts.source)));
    if (!kernel2) {
      LOG(ERROR) << "ProductKernel::Read: read failed: " << opts.source;
      return 0;
    }
    impl->kernel1_ = kernel1;
    impl->kernel2_ = kernel2;
    impl->Init();
    return impl;
  }

  bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    KernelHeader hdr;
    KernelImpl<O>::WriteHeader(strm, opts, kFileVersion, &hdr);
    if (!strm)
      return false;
    kernel1_->Write(strm, KernelWriteOptions(opts.source));
    kernel2_->Write(strm, KernelWriteOptions(opts.source));
    if (!strm)
      LOG(ERROR) << "ProductKernel::Write: write failed: " << opts.source;
    return strm;
  }

 private:
  void Init() {
    if (kernel1_->Data() == kernel2_->Data())
      KernelImpl<O>::SetData(kernel1_->Data());
    else
      LOG(FATAL) << "ProductKernel: incompatible datasets";
  }

  ProductKernelImpl() : kernel1_(0), kernel2_(0) {
    KernelImpl<O>::SetType("product");
  }

 private:
  Kernel<O> *kernel1_;
  Kernel<O> *kernel2_;

  // Identifies current file format version
  static const int32 kFileVersion = 1;
  // Identifies minimum compatible file format version
  static const int32 kMinFileVersion = 1;
};


template <class O>
class ProductKernel :public Kernel<O> {
 public:
  typedef O Object;

  ProductKernel(const Kernel<O> &k1, const Kernel<O> &k2) :
      impl_(new ProductKernelImpl<O>(k1, k2)) {

  }

  ProductKernel(const ProductKernel<O> &kernel) : impl_(kernel.impl_) {
    impl_->IncrRefCount();
  }

  virtual ~ProductKernel() { if(!impl_->DecrRefCount()) delete impl_;}

  ProductKernel<O> &operator=(const ProductKernel<O> &kernel) {
    if (this != &kernel) {
      if(!impl_->DecrRefCount()) delete impl_;
      kernel.impl_->IncrRefCount();
      impl_ = kernel.impl_;
    }
    return *this;
  }

  virtual double Eval(size_t i, size_t j) {
    return impl_->Eval(i,j);
  }

  virtual const string& Type() const { return impl_->Type(); }

  virtual ProductKernel<O> *Copy() const {
    impl_->IncrRefCount();
    return new ProductKernel<O>(impl_);
  }

  // Read and Write
  static ProductKernel<O> *Read(istream &strm,
                                   const KernelReadOptions &opts) {
    ProductKernelImpl<O> *impl = ProductKernelImpl<O>::Read(strm, opts);
    return impl ? new ProductKernel(impl) : 0;
  }

  static ProductKernel<O> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "ProductKernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

  virtual bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    impl_->Write(strm, opts);
    return strm;
  }

  virtual bool Write(const string &filename) const {
    if (!filename.empty()) {
      ofstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "ProductKernel::Write: Can't open file: " << filename;
        return false;
      }
      return Write(strm, KernelWriteOptions(filename));
    } else {
      return Write(std::cout, KernelWriteOptions("standard output"));
    }
  }

  virtual void SetData(DataSet<O>* ds) {
    impl_->SetData(ds);
  }

  virtual const DataSet<O>* Data() const {
    return impl_->Data();
  }

  virtual DataSet<Object>* Data() {
    return impl_->Data();
  }

  virtual uint64 Properties(uint64 mask) const {
    return impl_->Properties(mask);
  }

  virtual void SetProperties(uint64 props, uint64 mask) {
    impl_->SetProperties(props, mask);
  }

 protected:
  explicit ProductKernel(ProductKernelImpl<O> *impl) : impl_(impl) {}

 private:
  ProductKernelImpl<O> *impl_;
};

}  // namespace kernel

#endif  // KERNEL_LIB_PRODUCT_KERNEL_H__
